#include <stdlib.h>
#include <assert.h>

#include "param.h"
#include "family.h"
#include "coem.h"

#include <math.h>
#include <talloc.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_math.h>

coem* coem_create(family*       ef,
                  unsigned int  k,
                  unsigned int  s,
                  double**      data,
                  unsigned int* n
                  ) {
  coem* coem = malloc(sizeof(struct coem));
  coem->ef         = ef;
  coem->k          = k;
  coem->data       = data;
  coem->n          = n;
  coem->s          = s;
  coem->n_max      = 0;
  coem->n_total    = 0;
  coem->verbose    = 0;

  for(unsigned int i=0; i < s; i++) {
    coem->n_total += n[i];
    if (n[i] > coem->n_max)
      coem->n_max = n[i];
  }
  coem->mask       = malloc(coem->n_max * sizeof(char));

  coem->mixtures   = malloc(s * sizeof(mixture*));
  coem->weights    = malloc(s * sizeof(double*));
  coem->posteriors = malloc(s * sizeof(gsl_matrix*));
  coem->tmp_params = malloc(s * sizeof(param*));
  
  for(unsigned int l=0; l < s; l++) {
    coem->mixtures[l]   = mixture_create(k);
    coem->weights[l]    = coem->mixtures[l]->weights;
    coem->posteriors[l] = gsl_matrix_alloc(n[l], k);
  }
  coem->components = coem->mixtures[0]->params;

  /* share components between mixtures */
  for(unsigned int l=1; l < s; l++) {
    talloc_free(coem->mixtures[l]->params);
    coem->mixtures[l]->params = coem->mixtures[0]->params;
  }

  coem->treshold   = 1e-5;
  coem->maxiter    = 100;
  coem->iter       = 0;

  for(unsigned int l=0; l < s; l++) {
    for (int i=0; i<k; i++) {
      coem->mixtures[l]->ef[i] = ef;
    }
  }

  for (int i=0; i<k; i++) {
    coem->components[i] = (param*)ef->new_param(ef, SOURCE);
  }

  for (int i=0; i<s; i++) {
    coem->tmp_params[i] = (param*)ef->new_param(ef, EXPECTATION);
  }

  return coem;
}

void coem_initialize_from_coclustering(coem* coem, double **weights, unsigned int **affectation) {
  /* for each cluster */
  for (unsigned int i=0; i<coem->k; i++) {

    /* for each dataset */
    for (unsigned int l=0; l<coem->s; l++) {

      /* for each observation */
      unsigned int count = 0;
      for (unsigned int j=0; j<coem->n[l]; j++) {
        if (affectation[l][j] == i) {
          coem->mask[j] = 1;
          count++;
        }
        else
          coem->mask[j] = 0;
      }

      coem->ef->zero(coem->ef, coem->tmp_params[l]);

      if (count)
        coem->ef->mle(coem->ef, coem->n[l], coem->data[l], coem->mask, coem->tmp_params[l]);
      coem->weights[l][i] = weights[l][i];
    }
    coem->ef->zero(coem->ef, coem->components[i]);
    coem_aggregate(coem, coem->tmp_params, coem->components[i]);
  }
}

void coem_aggregate(coem* coem, param** src, param* dest) {
  family* ef = coem->ef;
  param* tmp = (param*)ef->new_param(ef, EXPECTATION);

  ef->zero(ef, tmp);
  for (unsigned int l=0; l<coem->s; l++) {
    ef->add(ef, tmp, src[l], tmp);
  }
  ef->scale(ef, tmp, 1./coem->s, tmp);

  ef->eta2lambda(ef, tmp, dest);
  ef->delete_param(ef, tmp);
}

void coem_expectation(coem* coem) {
  if (coem->verbose)
    printf("coem E\n");

  family*     ef         = coem->ef;
  param**     components = coem->components;
  double*     data;
  double*     weights;
  gsl_matrix* posterior;

  /* for each dataset */
#pragma omp parallel for private(data, weights, posterior)
  for (unsigned int l=0; l<coem->s; l++) {
    data      = coem->data[l];
    weights   = coem->weights[l];
    posterior = coem->posteriors[l];

    /* for each observation */
    for(unsigned int i=0; i<coem->n[l]; i++) {
      double sum = 0;

      /* for each cluster */
      for(unsigned int j=0; j<coem->k; j++) {
        double tmp = weights[j] * ef->pdf(ef, &data[i], components[j]);
        if (gsl_isnan(tmp))
          tmp = 0;
        
        sum += tmp;
        gsl_matrix_set(posterior, i, j, tmp);
      }

      /* for each cluster */
      for(unsigned int j=0; j<coem->k; j++) {
        double tmp = gsl_matrix_get(posterior, i, j);
        gsl_matrix_set(posterior, i, j, tmp / sum);
      }
    }
  }
}

void coem_maximization(coem* coem) {
  if (coem->verbose)
    printf("coem M\n");

  family* ef = coem->ef;
  unsigned int n;
  double     *data;
  double     *weights;
  gsl_matrix *posterior;

  /* for each cluster */
  for(unsigned int i=0; i<coem->k; i++) {

    /* for each dataset */
#pragma omp parallel for private(n, data, weights, posterior)
    for(unsigned int l=0; l<coem->s; l++) {
      param* eta = (param*)ef->new_param(ef, EXPECTATION);
      ef->zero(ef, eta);

      n         = coem->n[l];
      data      = coem->data[l];
      weights   = coem->weights[l];
      posterior = coem->posteriors[l];

      coem->ef->zero(coem->ef, coem->tmp_params[l]);

      /* for each observation */
      double sum = 0;
      for(unsigned int j=0; j<n; j++) {
        sum += gsl_matrix_get(posterior, j, i);
      }

      if (sum > 0) {
        /* for each observation */
        for(unsigned int j=0; j<n; j++) {
          double scale = gsl_matrix_get(posterior, j, i);
          ef->t(ef, &data[j], eta);
          ef->scale(ef, eta, scale, eta);
          ef->add(ef, coem->tmp_params[l], eta, coem->tmp_params[l]);
        }
        weights[i] = sum / n;
        ef->scale(ef, coem->tmp_params[l], 1. / sum, coem->tmp_params[l]);
      }

      ef->delete_param(ef, eta);
    }
    coem_aggregate(coem, coem->tmp_params, coem->components[i]);
  }

}

mixture** coem_run(coem* coem) {
  double old_ll = 0;

  coem->ll = 0;
  double ll = 0;
#pragma omp parallel for reduction(+:ll)
  for(unsigned int l=0; l<coem->s; l++) {
    ll += mixture_loglikelihood(coem->mixtures[l], coem->n[l], coem->data[l]);
  }
  coem->ll = ll;

  while (1) {
    if (coem->maxiter > 0 && coem->iter >= coem->maxiter)
      break;

    if (coem->treshold > 0 && fabs((old_ll - coem->ll) / old_ll) < coem->treshold)
      break;

    old_ll = coem->ll;
    coem_step(coem);
  }

  return coem->mixtures;
}

void coem_step(coem* coem) {
  if (coem->verbose)
    printf("Step %i\n", coem->iter);

  coem_expectation(coem);
  coem_maximization(coem);

  coem->ll = 0;
  double ll = 0;
#pragma omp parallel for reduction(+:ll)
  for(unsigned int l=0; l<coem->s; l++) {
    ll += mixture_loglikelihood(coem->mixtures[l], coem->n[l], coem->data[l]);
  }
  coem->ll = ll;

  coem->iter++;
}

void coem_destroy(coem* coem) {
  free(coem->mask);
  for(unsigned int l=0; l<coem->s; l++) {
    gsl_matrix_free(coem->posteriors[l]);
  }
  free(coem->posteriors);
  free(coem);
}

